"""
Basic framework for Pathspider: coordinate active measurements on large target
lists with both system-level network stack state (sysctls, iptables rules, etc)
as well as information derived from flow-level passive observation of traffic at
the sender.

.. moduleauthor:: Brian Trammell <brian@trammell.ch>

"""

import sys
import time
import logging
import collections
import threading
import multiprocessing as mp
import queue
from datetime import datetime

from pathspider.network import ipv4_address
from pathspider.network import ipv6_address
from pathspider.network import ipv4_address_public
from pathspider.network import ipv6_address_public
from pathspider.network import ipv4_asn
from pathspider.network import ipv6_asn

__version__ = "2.0.1"

CONN_OK = 0
CONN_FAILED = 1
CONN_TIMEOUT = 2
CONN_SKIPPED = 3
CONN_DISCARD = 4

PORT_FAILED = 0
PORT_FAILED_AGAIN = -1

QUEUE_SIZE = 1000
QUEUE_SLEEP = 0.5

SHUTDOWN_SENTINEL = "SHUTDOWN_SENTINEL"
NO_FLOW = None

class Spider:
    """
    A spider consists of a configurator (which alternates between two system
    configurations), a large number of workers (for performing some network
    action for each configuration), an Observer which derives information from
    passively observed traffic, and a thread that merges results from the
    workers with flow records from the collector.

    """

    chains = [] # Disable the observer by default

    def __init__(self, worker_count, libtrace_uri, args, server_mode):
        """
        The initialisation of a pathspider plugin.

        :param worker_count: The number of workers to use.
        :type worker_count: int
        :param libtrace_uri: The URI to pass to the Observer to describe the
                             interface on which packets should be captured.
        :type libtrace_uri: str

        It is expected that this function will be overloaded by plugins, though
        the plugin should always make a call to the __init__() function of the
        abstract Spider class as this initialises all of the base functionality:

        .. code-block:: python

         super().__init__(worker_count=worker_count,
                          libtrace_uri=libtrace_uri,
                          args=args)

        This can be used to initialise any variables which may be required in
        the object.
        """

        self.args = args

        self.running = False
        self.stopping = False
        self.terminating = False
        self.server_mode = server_mode

        self.worker_count = worker_count
        self.active_worker_count = 0
        self.active_worker_lock = threading.Lock()

        self.libtrace_uri = libtrace_uri

        self.jobqueue = queue.Queue(QUEUE_SIZE)
        self.resqueue = queue.Queue(QUEUE_SIZE)
        self.flowqueue = mp.Queue(QUEUE_SIZE)
        self.observer_shutdown_queue = mp.Queue(QUEUE_SIZE)

        self.jobtab = {}
        self.comparetab = {}
        self.restab = {}
        self.flowtab = {}
        self.flowreap = collections.deque()
        self.flowreap_size = min(self.worker_count * 100, 10000)

        self.outqueue = queue.Queue(QUEUE_SIZE)

        self.observer = None

        self.worker_threads = []
        self.configurator_thread = None
        self.merger_thread = None

        self.observer_process = None

        self.lock = threading.Lock()
        self.exception = None

        if libtrace_uri.startswith('int'):
            # TODO: Refactor this
            self.source = (ipv4_address(self.libtrace_uri[4:]),
                           ipv6_address(self.libtrace_uri[4:]))
            self.source_public = (ipv4_address_public(self.libtrace_uri[4:]),
                                  ipv6_address_public(self.libtrace_uri[4:]))
            self.source_asn = (ipv4_asn(self.libtrace_uri[4:]),
                               ipv6_asn(self.libtrace_uri[4:]))
        else:
            self.source = ("127.0.0.1", "::1")

        self.__logger = logging.getLogger('pathspider')

        if hasattr(self.args, 'connect') and self.args.connect.startswith('tor'):
            logging.getLogger("stem").setLevel(logging.ERROR)
            import stem.control
            self.controller = stem.control.Controller.from_port()
            self.controller.authenticate()


    def configurator(self):
        raise NotImplementedError("Cannot instantiate an abstract Spider")

    def worker(self, worker_number):
        raise NotImplementedError("Cannot instantiate an abstract Spider")

    def pre_connect(self, job):
        """
        Performs pre-connection operations.

        :param job: The job record
        :type job: dict

        The pre_connect function can be used to perform any operations that
        must be performed before each connection. It will be run only once
        per job, with the same result passed to both the A and B connect
        calls. This function is not synchronized with the configurator.

        Plugins to PATHspider can optionally implement this function. If this
        function is not overloaded, it will be a noop.
        """

        pass

    def _connect_wrapper(self, job, config, connect=None):
        start = str(datetime.utcnow())
        if connect is None:
            conn = self.connect(job, config) # pylint: disable=no-member
        else:
            if not hasattr(connect, '__self__'):
                connect = connect.__get__(self)
            conn = connect(job, config)
        conn['spdr_start'] = start
        return conn

    def post_connect(self, job, rec, config):
        """
        Performs post-connection operations.

        :param job: The job record.
        :type job: dict
        :param rec: The result of the connection operation(s).
        :type rec: dict
        :param config: The state of the configurator during
                       :func:`pathspider.base.Spider.connect`.
        :type config: int

        The post_connect function can be used to perform any operations that
        must be performed after each connection. It will be run for both the
        A and the B configuration, and is not synchronized with the
        configurator.

        Plugins to PATHspider can optionally implement this function. If this
        function is not overloaded, it will be a noop.

        Any sockets or other file handles that were opened during
        :func:`pathspider.base.Spider.connect` should be closed in this
        function if they have not been already.
        """

        pass

    def create_observer(self):
        """
        Create a flow observer.

        This function is called by the base Spider logic to get an instance
        of :class:`pathspider.observer.Observer` configured with the function
        chains that are requried by the plugin.
        """

        self.__logger.info("Creating observer")
        if len(self.chains) > 0:
            from pathspider.observer import Observer
            return Observer(self.libtrace_uri,
                            chains=self.chains) # pylint: disable=no-member
        else:
            from pathspider.observer import DummyObserver
            return DummyObserver()

    def _key(self, obj):
        if self.server_mode:
            objkey = (obj['sip'], obj['sp'])
        else:
            objkey = (obj['dip'], obj['sp'])
        return objkey

    def _merge_flows(self):
        try:
            flow = self.flowqueue.get_nowait()
        except queue.Empty:
            time.sleep(QUEUE_SLEEP)
            return True
        else:
            if flow == SHUTDOWN_SENTINEL:
                self.__logger.debug("stopping flow merging on sentinel")
                return False

            flowkey = self._key(flow)
            self.__logger.debug("got a flow (" + repr(flowkey) + ")")

            if flowkey in self.restab:
                self.__logger.debug("merging flow")
                self.merge(flow, self.restab[flowkey])
                del self.restab[flowkey]
            elif flowkey in self.flowtab:
                self.__logger.debug("won't merge duplicate flow")
            else:
                # Create a new flow
                self.flowtab[flowkey] = flow

                # And reap the oldest, if the reap queue is full
                self.flowreap.append(flowkey)
                if len(self.flowreap) > self.flowreap_size:
                    try:
                        del self.flowtab[self.flowreap.popleft()]
                    except KeyError:
                        pass
            return True

    def _merge_results(self):
        try:
            res = self.resqueue.get_nowait()
        except queue.Empty:
            time.sleep(QUEUE_SLEEP)
            self.__logger.debug("result queue is empty")
            return True
        else:
            if res == SHUTDOWN_SENTINEL:
                self.__logger.debug("stopping result merging on sentinel")
                return False
            if 'spdr_state' in res.keys() and res['spdr_state'] == CONN_SKIPPED:
                # handle skipped results
                return True

            reskey = self._key(res)
            self.__logger.debug("got a result (" + repr(reskey) + ")")

            if reskey in self.restab and res['sp'] == PORT_FAILED:
                # both connections failed, but need to be distinguished
                reskey = (reskey[0], PORT_FAILED_AGAIN)

            if reskey in self.flowtab:
                self.__logger.debug("merging result")
                self.merge(self.flowtab[reskey], res)
                del self.flowtab[reskey]
            elif reskey in self.restab:
                self.__logger.debug("won't merge duplicate result")
            else:
                self.restab[reskey] = res

            self.resqueue.task_done()
            return True

    def merger(self):
        """
        Thread to merge results from the workers and the observer.
        """

        if len(self.chains) == 0:
            self.__logger.warning("Merger is not expecting flows from the Observer")
            # Immediately merge with NO_FLOW when there's no chains, as there's
            # going to also be no observer and so no flows.
            while self.running:
                try:
                    res = self.resqueue.get_nowait()
                except queue.Empty:
                    time.sleep(QUEUE_SLEEP)
                    self.__logger.debug("result queue is empty")
                    continue
                if res == SHUTDOWN_SENTINEL:
                    break
                if not ('spdr_state' in res.keys() and res['spdr_state'] == CONN_SKIPPED):
                    self.merge(NO_FLOW, res)
        else:
            merging_flows = True
            merging_results = True

            while self.running and (merging_results or merging_flows):

                if merging_flows and self.flowqueue.qsize() >= self.resqueue.qsize():
                    merging_flows = self._merge_flows()

                elif merging_results:
                    merging_results = self._merge_results()

        # One more pass during shutdown, to clean up any leftovers
        for res_item in self.restab.items():
            res = res_item[1]
            self.merge(NO_FLOW, res)

    def merge(self, flow, res):
        """
        Merge a job record with a flow record.

        :param flow: The flow record.
        :type flow: dict
        :param res: The job record.
        :type res: dict
        :return: tuple -- Final record for job.

        In order to create a final record for reporting on a job, the final job
        record must be merged with the flow record. This function should
        be implemented by any plugin to provide the logic for this merge as
        the keys used in these records cannot be known by PATHspider in advance.

        This method is not implemented in the abstract
        :class:`pathspider.base.Spider` class and must be implemented by any
        plugin.
        """

        if flow == NO_FLOW:
            flow = {'observed': False}
        else:
            flow['observed'] = True

        for key in res.keys():
            if key in flow.keys():
                if res[key] == flow[key]:
                    continue
                else:
                    self.__logger.warning("Dropping flow due to mismatch with "
                                          "observations on key %s", key)
                    return
            flow[key] = res[key]

        # Remove private keys - we need to make a copy of the keys as we
        #                       modify the dict during the iteration
        for key in [x for x in flow.keys()]:
            if key.startswith("_"):
                flow.pop(key)

        self.__logger.debug("Result: " + str(flow))

        if flow['jobId'] not in self.comparetab:
            self.comparetab[flow['jobId']] = []
        self.comparetab[flow['jobId']].append(flow)

        if len(self.comparetab[flow['jobId']]) == self._config_count: # pylint: disable=no-member
            flows = self.comparetab.pop(flow['jobId'])
            flows.sort(key=lambda x: x['config'])
            start = min([flow['spdr_start'] for flow in flows])
            stop = max([flow['spdr_stop'] for flow in flows])
            job = self.jobtab.pop(flow['jobId'])
            job['flow_results'] = flows
            job['time'] = {'from': start, 'to': stop}
            job['missed_flows'] = 0
            for flow in flows:
                if not flow['observed']:
                    job['missed_flows'] = job['missed_flows'] + 1
            job['conditions'] = self.combine_flows(flows)
            if job['conditions'] is not None:
                if "pathspider.not_observed" in job['conditions']:
                    self.__logger.debug("At least one flow was not observed and so conditions could not be fully generated (if at all)")
                if job['missed_flows'] > 0:
                    job['conditions'].append("pathspider.missed_flows:" + str(job['missed_flows']))
            else:
                job.pop('conditions')
            self.outqueue.put(job)

    def combine_flows(self, flows):
        pass

    def exception_wrapper(self, target, *args, **kwargs):
        try:
            target(*args, **kwargs)
        except: # pylint: disable=W0702
            self.__logger = logging.getLogger('pathspider')
            self.__logger.exception("exception occurred. terminating.")
            if self.exception is None:
                self.exception = sys.exc_info()[1]

            self.terminate()

    def _finalise_conns(self, job, jobId, conns):
        # Pass results on for merge
        config = 0
        for conn in conns:
            self.post_connect(job, conn, config)
            conn['spdr_stop'] = str(datetime.utcnow())
            conn['config'] = config
            if self.server_mode:
                conn['sip'] = job['sip']
            else:
                conn['dip'] = job['dip']
            conn['jobId'] = jobId
            self.resqueue.put(conn)
            config += 1

    def start(self):
        """
        This function starts a PATHspider plugin by:

         * Setting the running flag
         * Create and start an observer
         * Start the merger thread
         * Start the configurator thread
         * Start the worker threads

        The number of worker threads to start was given when activating the
        plugin.
        """

        self.__logger.info("starting pathspider")

        with self.lock:
            # set the running flag
            self.running = True

            # create an observer and start its process
            self.observer = self.create_observer()
            self.observer_process = mp.Process(
                args=(self.observer.run_flow_enqueuer,
                      self.flowqueue,
                      self.observer_shutdown_queue),
                target=self.exception_wrapper,
                name='observer',
                daemon=True)
            self.observer_process.start()
            self.__logger.debug("observer forked")

            # now start up ecnspider, backwards
            self.merger_thread = threading.Thread(
                args=(self.merger,),
                target=self.exception_wrapper,
                name="merger",
                daemon=True)
            self.merger_thread.start()
            self.__logger.debug("merger up")

            self.configurator_thread = threading.Thread(
                args=(self.configurator,),
                target=self.exception_wrapper,
                name="configurator",
                daemon=True)
            self.configurator_thread.start()
            self.__logger.debug("configurator up")
            self.worker_threads = []
            with self.active_worker_lock:
                self.active_worker_count = self.worker_count
            for i in range(self.worker_count):
                worker_thread = threading.Thread(
                    args=(self.worker, i),
                    target=self.exception_wrapper,
                    name='worker_{}'.format(i),
                    daemon=True)
                self.worker_threads.append(worker_thread)
                worker_thread.start()
            self.__logger.debug("workers up")

    def shutdown(self):
        """
        Shut down PathSpider in an orderly fashion,
        ensuring that all queued jobs complete,
        and all available results are merged.

        """

        self.__logger.info("beginning shutdown")

        with self.lock:
            # Set stopping flag
            self.stopping = True

            # Put a bunch of shutdown signals in the job queue
            for _ in range(self.worker_count * 2):
                self.jobqueue.put(SHUTDOWN_SENTINEL)

            # Wait for worker threads to shut down
            for worker in self.worker_threads:
                if threading.current_thread() != worker:
                    self.__logger.debug("joining worker: " + repr(worker))
                    worker.join()
            self.__logger.debug("all workers joined")

            # Tell observer to shut down
            self.observer_shutdown_queue.put(True)
            self.observer_process.join()
            self.__logger.debug("observer shutdown")

            # Tell merger to shut down
            self.resqueue.put(SHUTDOWN_SENTINEL)
            self.merger_thread.join()
            self.__logger.debug("merger shutdown")

            # Wait for merged results to be written
            self.outqueue.join()
            self.__logger.debug("all results retrieved")

            # Propagate shutdown sentinel and tell threads to stop
            self.outqueue.put(SHUTDOWN_SENTINEL)

            # Tell threads we've stopped
            self.running = False

            # Join configurator
            # if threading.current_thread() != self.configurator_thread:
            #     self.configurator_thread.join()

            self.stopping = False

        self.__logger.info("shutdown complete")

    def terminate(self):
        """
        Shut down PathSpider as quickly as possible,
        without any regard to completeness of results.

        """
        self.__logger.info("terminating pathspider")

        # tell threads to stop
        self.stopping = True
        self.running = False

        # terminate observer
        self.observer_shutdown_queue.put(True)

        # drain queues
        try:
            while True:
                self.jobqueue.task_done()
        except ValueError:
            pass

        try:
            while True:
                self.resqueue.task_done()
        except ValueError:
            pass

        try:
            while True:
                self.flowqueue.get_nowait()
        except queue.Empty:
            pass

        # Join remaining threads
        for worker in self.worker_threads:
            if threading.current_thread() != worker:
                self.__logger.debug("joining worker: " + repr(worker))
                worker.join()
        self.__logger.debug("all workers joined")

        if self.configurator_thread and \
                (threading.current_thread() != self.configurator_thread):
            self.configurator_thread.join()
            self.__logger.debug("configurator joined")

        if threading.current_thread() != self.merger_thread:
            self.merger_thread.join()
            self.__logger.debug("merger joined")

        self.observer_process.join()
        self.__logger.debug("observer joined")

        self.outqueue.put(SHUTDOWN_SENTINEL)
        self.__logger.info("termination complete")

    def add_job(self, job):
        """
        Adds a job to the job queue.

        If PATHspider is currently stopping, the job will not be added to the
        queue.
        """

        if self.stopping:
            return

        if not self.server_mode:
            if 'dip' in job.keys():
                sourceindex = 1 if ':' in job['dip'] else 0
                job['sip'] = self.source[sourceindex]
                job['path'] = [job['sip']]
                job['sip_public'] = self.source_public[sourceindex]
                if not ( job['sip'] == job['sip_public'] ):
                    job['path'].append(job['sip_public'])
                if self.source_asn[sourceindex] is not None:
                    job['sip_asn'] = self.source_asn[sourceindex]
                    job['path'].append("AS" + str(job['sip_asn']))
                if 'dip_asn' in job.keys(): # This may be generated by other tools
                    job['path'].append("AS" + job['dip_asn'])
                elif 'info' in job.keys(): # Hellfire does it this way
                    if 'ASN' in job['info'].keys():
                        job['path'].append("AS" + str(job['info']['ASN']))
                job['path'].append(job['dip'])

        self.jobqueue.put(job)

class PluggableSpider:
    @staticmethod
    def register_args(subparsers):
        raise NotImplementedError("Cannot register an abstract plugin")



"""
.. module:: pathspider.chains.dscp
   :synopsis: A flow analysis chain for UDP

This module contains the UDPChain flow analysis chain which can be used by
PATHspider's Observer for recording UDP details.

.. codeauthor:: Iain R. Learmonth <irl@fsfe.org>

"""

from pathspider.chains.base import Chain

class UDPChain(Chain):
    """

    +---------------------------+------+---------------------------------------+
    | Field Name                | Type | Meaning                               |
    +===========================+======+=======================================+
    | ``udp_zero_checksum_fwd`` | bool | True if the last packet in the flow   |
    |                           |      | in the forward direction had the UDP  |
    |                           |      | checksum disabled (set to zero).      |
    +---------------------------+------+---------------------------------------+
    | ``udp_zero_checksum_rev`` | bool | True if the last packet in the flow   |
    |                           |      | in the reverse direction had the UDP  |
    |                           |      | checksum disabled (set to zero).      |
    +---------------------------+------+---------------------------------------+
    """

    def new_flow(self, rec, ip):
        """
        For a new flow, all fields will be initialised to ``None``.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IP or IPv6 packet that triggered the creation of a new
                   flow record
        :type ip: plt.ip or plt.ip6
        :return: Always ``True``
        :rtype: bool
        """

        rec['udp_zero_checksum_fwd'] = None
        rec['udp_zero_checksum_rev'] = None
        return True

    def udp(self, rec, udp, rev):
        """
        Records details from UDP datagram about the UDP header.

        :param rec: the flow record
        :type rec: dict
        :param tcp: the UDP packet that was observed to be part of this flow
        :type ip: plt.udp
        :param rev: ``True`` if the packet was in the reverse direction, ``False`` if
                    in the forward direction
        :type rev: bool
        :return: Always ``True``
        :rtype: bool
        """

        rec['udp_zero_checksum_rev' if rev else 'udp_zero_checksum_fwd'] = udp.checksum == 0

        return True



from pathspider.chains.mss import MSSChain
from pathspider.plugins.mss import MSS
from pathspider.tests.chains import ChainTestCase
from pathspider.chains.tcp import TCP_SA 

def test_plugin_mss_combine_not_observed():
    flows = [
             {'observed': False, 'tcp_connected': False},
             {'observed': True, 'tcp_connected': False}
            ]
    conditions = MSS.combine_flows(None, flows)
    assert "pathspider.not_observed" in conditions


def test_plugin_mss_combine():
    test_groups = [
                   (True,  "mss.connectivity.online"),
                   (False, "mss.connectivity.offline")
                  ]
    for group in test_groups:
        flows = [              
                 {'observed': True, 'tcp_connected': group[0], 'mss_value_fwd': None , 'mss_len_rev': None, 'mss_value_rev': None}
                ]
        
        conditions = MSS.combine_flows(None, flows)
        assert group[1] in conditions
        if "mss.connectivity.online" in conditions:
            assert 'mss.option.received.absent' in conditions
         

    mss_groups = [ 
                  ('mss.option.received.unchanged', 1460, 1460),
                  ('mss.option.received.inflated', 1440, 1460),
                  ('mss.option.received.deflated', 1460, 1440)
                 ]
    for group in mss_groups:
        flow =  [
                 {'observed': True, 'tcp_connected': True, 'mss_value_fwd': group[1], 'mss_len_rev': 4, 'mss_value_rev': group[2]}
                ]
        conditions = MSS.combine_flows(None, flow)
        assert group[0] in conditions



"""
.. module:: pathspider.chains.tcp
   :synopsis: A flow analysis chain for TCP Maximum Segment Size

This module contains the MSSChain flow analysis chain which can be used by
PATHspider's Observer for recording TCP Maximum Segment Size details.

.. codeauthor:: Iain R. Learmonth <irl@fsfe.org>

"""

from pathspider.chains.base import Chain
from pathspider.chains.tcp import tcp_options
from pathspider.chains.tcp import TO_MSS

class MSSChain(Chain):
    """
    This flow analysis chain records details of the TCP Maximum Segment Size
    (MSS) option in the flow record. It will determine the length and value of
    the field if present in SYN packets.

    +------------------+--------+-----------------------------------------------------------------+
    | Field Name       | Type   | Meaning                                                         |
    +==================+========+=================================================================+
    | ``mss_len_fwd``  | int    | Length of the MSS option field including kind and length in the |
    |                  |        | forward direction.                                              |
    +------------------+--------+-----------------------------------------------------------------+
    | ``mss_len_rev``  | int    | Length of the MSS option field including kind and length in the |
    |                  |        | reverse direction.                                              |
    +------------------+--------+-----------------------------------------------------------------+
    | ``mss_value_fwd``| int    | Value of the MSS option field in the forward direction.         |
    +------------------+--------+-----------------------------------------------------------------+
    | ``mss_value_rev``| int    | Value of the MSS option field in the reverse direction.         |
    +------------------+--------+-----------------------------------------------------------------+
    """

    def new_flow(self, rec, ip):
        """
        For a new flow, all fields will be initialised to ``None``.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IP or IPv6 packet that triggered the creation of a new
                   flow record
        :type ip: plt.ip or plt.ip6
        :return: Always ``True``
        :rtype: bool
        """

        rec['mss_len_fwd'] = None
        rec['mss_len_rev'] = None
        rec['mss_value_fwd'] = None
        rec['mss_value_rev'] = None

        return True

    def tcp(self, rec, tcp, rev): # pylint: disable=unused-argument
        """
        Records TCP Maximum Segment Size Details.

        TCP Maximum Segment Size
            The TCP options will be parsed for the MSS option for all SYN
            packets.  If the option is found, the length and value for the
            option will be recorded in the flow.

        :param rec: the flow record
        :type rec: dict
        :param tcp: the TCP segment that was observed to be part of this flow
        :type ip: plt.tcp
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: Always True
        :rtype: bool
        """

        # Shortcut non-SYN
        if not tcp.syn_flag:
            return True

        opts = tcp_options(tcp)

        if TO_MSS in opts:
            mss = bytes(opts[TO_MSS])
            rec['mss_len_' + ('rev' if rev else 'fwd')] = len(mss) + 2
            rec['mss_value_' + ('rev' if rev else 'fwd')] = int.from_bytes(mss, byteorder="big")

        # tell observer to keep going
        return True




import pycurl

import pathspider.base
from pathspider.base import PluggableSpider
from pathspider.base import CONN_OK
from pathspider.desync import DesynchronizedSpider
from pathspider.helpers.http import connect_http
from pathspider.helpers.http import connect_https
from pathspider.helpers.tor_http import connect_tor_http
from pathspider.chains.basic import BasicChain
from pathspider.chains.tcp import TCPChain

class H2(DesynchronizedSpider, PluggableSpider):

    name = "h2"
    description = "HTTP/2"
    version = pathspider.base.__version__
    chains = [BasicChain, TCPChain]
    connect_supported = ["http", "https", "torhttp"]

    def conn_no_h2(self, job, config):  # pylint: disable=unused-argument
        if self.args.connect == "http":
            return connect_http(self.source, job, self.args.timeout)
        if self.args.connect == "https":
            return connect_http(self.source, job, self.args.timeout)
        if self.args.connect == "torhttp":
            return connect_tor_http(self.controller, self.args.tor_path, job, self.args.timeout)
        else:
            raise RuntimeError("Unknown connection mode specified")

    def conn_h2(self, job, config): # pylint: disable=unused-argument
        curlopts = {pycurl.HTTP_VERSION: pycurl.CURL_HTTP_VERSION_2_0}
        curlinfos = {pycurl.INFO_HTTP_VERSION}
        if self.args.connect == "http":
            return connect_http(self.source, job, self.args.timeout, curlopts, curlinfos)
        if self.args.connect == "https":
            return connect_https(self.source, job, self.args.timeout, curlopts, curlinfos)
        if self.args.connect == "torhttp":
            return connect_tor_http(self.controller, self.args.tor_path, job, self.args.timeout, curlopts, curlinfos)
        else:
            raise RuntimeError("Unknown connection mode specified")

    connections = [conn_no_h2, conn_h2]

    def combine_flows(self, flows):
        conditions = []

        if (not flows[0]['spdr_state'] == CONN_OK and
                not flows[1]['spdr_state'] == CONN_OK):
            conditions.append('h2.connectivity.offline')
        elif (not flows[0]['spdr_state'] == CONN_OK and
              flows[1]['spdr_state'] == CONN_OK):
            conditions.append('h2.connectivity.transient')
        elif (flows[0]['spdr_state'] == CONN_OK and
              flows[1]['spdr_state'] == CONN_OK):
            conditions.append('h2.connectivity.works')
            if flows[1]['http_info'][pycurl.INFO_HTTP_VERSION] == pycurl.CURL_HTTP_VERSION_2_0:
                conditions.append('h2.upgrade.success')
            else:
                conditions.append('h2.upgrade.failed')
        else:
            conditions.append('h2.connectivity.broken')

        return conditions




import logging
import pathspider
import nose

def runtestsuite(args):
    # Undo the logging basic config to allow for log capture
    root_logger = logging.getLogger()
    for handler in list(root_logger.handlers):
        root_logger.removeHandler(handler)

    # Collect and run tests
    nose.run(argv=['nosetests', '-v', 'pathspider'])

def register_args(subparsers):
    parser = subparsers.add_parser(name='test',
                                   help="Run the built in test suite")

    # Set the command entry point
    parser.set_defaults(cmd=runtestsuite)




from collections import namedtuple
from tempfile import NamedTemporaryFile

from scapy.all import wrpcap

from pathspider.chains.evil import EvilChain
from pathspider.plugins.evilbit import EvilBit
from pathspider.tests.chains import ChainTestCase
from pathspider.chains.tcp import TCP_SA 

TestArgs = namedtuple('TestArgs', ['connect'])
job = {'dip': '192.0.2.4', 'dp': 80, 'domain': 'example.com'}

class TestPluginEvilBitForgeObserve(ChainTestCase):

    def test_plugin_evilbit_forge_observer(self):
        for connect in EvilBit.connect_supported:
            spider = EvilBit(0, "", TestArgs(connect=connect))
            packets = []

            for seq in range(0, spider.packets):
                packets.append(spider.forge(job, seq))
            assert len(packets) == 2

            expected_flows = [
              {
                'evilbit_syn_fwd': False if connect == "tcpsyn" else None,
                'evilbit_syn_rev': None,
                'evilbit_data_fwd': False if connect == "dnsudp" else None,
                'evilbit_data_rev': None
              }, {
                'evilbit_syn_fwd': True if connect == "tcpsyn" else None,
                'evilbit_syn_rev': None,
                'evilbit_data_fwd': True if connect == "dnsudp" else None,
                'evilbit_data_rev': None
              }
            ]

            with NamedTemporaryFile() as test_trace:
                for idx in range(0, 2):
                    wrpcap(test_trace.name, [packets[idx]])
                    self.create_observer(test_trace.name, [EvilChain])
        
                    flows = self.run_observer()
                    assert len(flows) == 1
        
                    for key in expected_flows[idx]:
                        print(key + ">>" + str(flows[0][key]) + ":" +
                              str(expected_flows[0][key]))
                        assert flows[0][key] == expected_flows[idx][key]


def test_plugin_evilbit_forge_diff():
    for connect in EvilBit.connect_supported:
        spider = EvilBit(0, "", TestArgs(connect=connect))
        packets = []

        for seq in range(0, spider.packets):
            packets.append(spider.forge(job, seq))
        assert len(packets) == 2

        packets[0].flags = 'evil'
        packets[0].payload.sport = packets[1].payload.sport

        print(packets[0].summary())
        print(packets[1].summary())

        assert bytes(packets[0]) == bytes(packets[1])

def test_plugin_evilbit_combine_not_observed():
    flows = [
             {'observed': True},
             {'observed': False}
            ]
    conditions = EvilBit.combine_flows(None, flows)
    assert "pathspider.not_observed" in conditions


def test_plugin_evilbit_combine():
    test_groups = [
                   (True,  True,  "evilbit.connectivity.works", TCP_SA, TCP_SA),
                   (True,  False, "evilbit.connectivity.broken", TCP_SA, None, ),
                   (False, True,  "evilbit.connectivity.transient", None, TCP_SA),
                   (False, False, "evilbit.connectivity.offline", None, None)
                  ]
    for group in test_groups:
        flows = [
                [
                 {'observed': True, 'tcp_synflags_rev': group[3]},
                 {'observed': True, 'tcp_synflags_rev': group[4], 'evilbit_syn_rev': True}
                ],
                [
                 {'observed': True, 'dns_response_valid': group[0]},
                 {'observed': True, 'dns_response_valid': group[1], 'evilbit_data_rev': True}
                ]]
        
        for flow in [0,1]:
            spider = EvilBit(0, "", TestArgs(connect=EvilBit.connect_supported[flow]))
            conditions = EvilBit.combine_flows(spider, flows[flow])
            assert group[2] in conditions


    evil_groups = [ ('evilbit.mark.seen', True),
                    ('evilbit.mark.not_seen', False)
                  ]
    for group in evil_groups:
        flows = [
                [
                 {'observed': True, 'tcp_synflags_rev': TCP_SA},
                 {'observed': True, 'tcp_synflags_rev': TCP_SA, 'evilbit_syn_rev': group[1]}
                ],
                [
                 {'observed': True, 'dns_response_valid': True},
                 {'observed': True, 'dns_response_valid': True, 'evilbit_data_rev': group[1]}
                ]]
        for flow in [0,1]:
            spider = EvilBit(0, "", TestArgs(connect=EvilBit.connect_supported[flow]))
            conditions = EvilBit.combine_flows(spider, flows[flow])
            assert group[0] in conditions

            




from pathspider.tests.chains import ChainTestCase

from pathspider.chains.ecn import ECNChain
from pathspider.chains.tcp import TCP_SEC
from pathspider.chains.tcp import TCP_SAE

class TestECNChain(ChainTestCase):

    def test_chain_ecn_ipv6_unreachable_ce_on_syn(self):
        test_trace = "ecn_ipv6_unreachable_ce_on_syn.pcap"
        self.create_observer(test_trace, [ECNChain])

        expected_ecn = {
            'ecn_ect0_syn_fwd': False,
            'ecn_ect1_syn_fwd': False,
            'ecn_ce_syn_fwd': True,
            'ecn_ect0_data_fwd': False,
            'ecn_ect1_data_fwd': False,
            'ecn_ce_data_fwd': False,
            'ecn_ect0_syn_rev': False,
            'ecn_ect1_syn_rev': False,
            'ecn_ce_syn_rev': False,
            'ecn_ect0_data_rev': False,
            'ecn_ect1_data_rev': False,
            'ecn_ce_data_rev': False,
        }
        
        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_ecn:
            assert flows[0][key] == expected_ecn[key]

    def test_chain_ecn_ipv4_nice_ect0(self):
        test_trace = "ecn_ipv4_nice_ect0.pcap"
        self.create_observer(test_trace, [ECNChain])

        expected_ecn = {
            'ecn_ect0_syn_fwd': False,
            'ecn_ect1_syn_fwd': False,
            'ecn_ce_syn_fwd': False,
            'ecn_ect0_data_fwd': True,
            'ecn_ect1_data_fwd': False,
            'ecn_ce_data_fwd': False,
            'ecn_ect0_syn_rev': False,
            'ecn_ect1_syn_rev': False,
            'ecn_ce_syn_rev': False,
            'ecn_ect0_data_rev': True,
            'ecn_ect1_data_rev': False,
            'ecn_ce_data_rev': False,
        }
        
        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_ecn:
            assert flows[0][key] == expected_ecn[key]

    def test_chain_ecn_ipv6_nice_ect0(self):
        test_trace = "ecn_ipv6_nice_ect0.pcap"
        self.create_observer(test_trace, [ECNChain])

        expected_ecn = {
            'ecn_ect0_syn_fwd': False,
            'ecn_ect1_syn_fwd': False,
            'ecn_ce_syn_fwd': False,
            'ecn_ect0_data_fwd': True,
            'ecn_ect1_data_fwd': False,
            'ecn_ce_data_fwd': False,
            'ecn_ect0_syn_rev': False,
            'ecn_ect1_syn_rev': False,
            'ecn_ce_syn_rev': False,
            'ecn_ect0_data_rev': True,
            'ecn_ect1_data_rev': False,
            'ecn_ce_data_rev': False,
        }
        
        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_ecn:
            assert flows[0][key] == expected_ecn[key]

    def test_chain_ecn_fake_fwd_ect1(self):
        test_trace = "ecn_fake_fwd_ect1.pcap"
        self.create_observer(test_trace, [ECNChain])

        expected_ecn = {
            'ecn_ect0_syn_fwd': False,
            'ecn_ect1_syn_fwd': True,
            'ecn_ce_syn_fwd': False,
            'ecn_ect0_data_fwd': False,
            'ecn_ect1_data_fwd': True,
            'ecn_ce_data_fwd': False,
            'ecn_ect0_syn_rev': False,
            'ecn_ect1_syn_rev': False,
            'ecn_ce_syn_rev': False,
            'ecn_ect0_data_rev': False,
            'ecn_ect1_data_rev': False,
            'ecn_ce_data_rev': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_ecn:
            assert flows[0][key] == expected_ecn[key]

    def test_chain_ecn_fake_fwd_ce(self):
        test_trace = "ecn_fake_fwd_ce.pcap"
        self.create_observer(test_trace, [ECNChain])

        expected_ecn = {
            'ecn_ect0_syn_fwd': False,
            'ecn_ect1_syn_fwd': False,
            'ecn_ce_syn_fwd': True,
            'ecn_ect0_data_fwd': False,
            'ecn_ect1_data_fwd': False,
            'ecn_ce_data_fwd': True,
            'ecn_ect0_syn_rev': False,
            'ecn_ect1_syn_rev': False,
            'ecn_ce_syn_rev': False,
            'ecn_ect0_data_rev': False,
            'ecn_ect1_data_rev': False,
            'ecn_ce_data_rev': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_ecn:
            assert flows[0][key] == expected_ecn[key]



"""
.. module:: pathspider.chains.tcp
   :synopsis: A flow analysis chain for TCP Fast Open

This module contains the TFOChain flow analysis chain which can be used by
PATHspider's Observer for recording TCP Fast Open [RFC7413]_ details.

.. warning: The names of fields used in this chain may be changed soon to
            more closely align with the field names used in other chains.

.. codeauthor:: Iain R. Learmonth <irl@fsfe.org>
.. codeauthor:: Piet De Vaere <piet@devae.re>

"""

from pathspider.chains.base import Chain
from pathspider.chains.tcp import tcp_options
from pathspider.chains.tcp import TO_FASTOPEN
from pathspider.chains.tcp import TO_EXID_FASTOPEN
from pathspider.chains.tcp import TO_EXPA
from pathspider.chains.tcp import TO_EXPB

class TFOChain(Chain):
    """
    This flow analysis chain records details of TCP Fast Open use in
    the flow record. It will determine whether the IANA assigned TCP option
    kind or the TCP Option Experiment ID [RFC6994]_ was used to identify the
    option, and whether the data sent on the SYN was acknowledged.

    +------------------+--------+-----------------------------------------------------------------+
    | Field Name       | Type   | Meaning                                                         |
    +==================+========+=================================================================+
    | ``tfo_synkind``  | int    | Identified by ``pathspider.chains.tcp.TO_{FASTOPEN,EXPA,EXPB}`` |
    +------------------+--------+-----------------------------------------------------------------+
    | ``tfo_ackkind``  | int    | Identified by ``pathspider.chains.tcp.TO_{FASTOPEN,EXPA,EXPB}`` |
    +------------------+--------+-----------------------------------------------------------------+
    | ``tfo_synclen``  | int    | TFO cookie length in the forward direction                      |
    +------------------+--------+-----------------------------------------------------------------+
    | ``tfo_ackclen``  | int    | TFO cookie length in the reverse direction                      |
    +------------------+--------+-----------------------------------------------------------------+
    | ``tfo_dlen``     | int    | Length of SYN payload in the forward direction                  |
    +------------------+--------+-----------------------------------------------------------------+
    | ``tfo_ack``      | int    | Bytes acknowledged on the SYN in the reverse direction          |
    +------------------+--------+-----------------------------------------------------------------+
    """

    def _cookie(self, tcp):
        opts = tcp_options(tcp)

        if TO_FASTOPEN in opts:
            return (TO_FASTOPEN, bytes(opts[TO_FASTOPEN]))
        elif TO_EXPA in opts and opts[TO_EXPA][0:2] == bytearray(TO_EXID_FASTOPEN):
            return (TO_EXPA, bytes(opts[TO_EXPA][2:]))
        elif TO_EXPB in opts and opts[TO_EXPB][0:2] == bytearray(TO_EXID_FASTOPEN):
            return (TO_EXPB, tuple(opts[TO_EXPA][2:]))
        else:
            return (None, None)

    def new_flow(self, rec, ip):
        """
        For a new flow, all fields will be initialised to ``int(0)``.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IP or IPv6 packet that triggered the creation of a new
                   flow record
        :type ip: plt.ip or plt.ip6
        :return: Always ``True``
        :rtype: bool
        """

        rec['tfo_synkind'] = 0
        rec['tfo_ackkind'] = 0
        rec['tfo_synclen'] = 0
        rec['tfo_ackclen'] = 0
        rec['tfo_seq'] = 0
        rec['tfo_dlen'] = 0
        rec['tfo_ack'] = 0

        return True

    def tcp(self, rec, tcp, rev): # pylint: disable=unused-argument
        """
        Records TCP Fast Open details.

        TCP Option Used
            The TCP options will be parsed for options that use either the
            IANA assigned TCP option number or one of the TCP Option Experiment
            option numbers with the TCP Option Experiment ID used by TCP Fast
            Open early in its standardisiation. If an option is found, the
            method by which it was identified will be recorded in the
            ``tfo_synkind`` field for the forward direction and ``tfo_ackkind``
            field for the reverse direction.

        TCP Fast Open Cookie Length
            The length of the cookies observed on TCP options will be recorded
            in the ``tfo_synclen`` field for the forward direction and
            ``tfo_ackclen`` for the reverse direction. If no Fast Open option
            is found, this will remain at 0 when the flow is complete.

        Acknowledgement of SYN data
            The length of the data on the SYN in the forward direction will be
            recorded in the ``tfo_dlen`` field. The TCP sequence number for the
            SYN in the forward direction will be recorded in ``tfo_seq`` field
            and the TCP acknowledgement number for the SYN in the reverse
            direction will be recorded in the ``tfo_ack`` field.

        :param rec: the flow record
        :type rec: dict
        :param tcp: the TCP segment that was observed to be part of this flow
        :type ip: plt.tcp
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: Always True
        :rtype: bool
        """

        # Shortcut non-SYN
        if not tcp.syn_flag:
            return True

        # Check for TFO cookie and data on SYN
        if tcp.syn_flag and not tcp.ack_flag:
            (tfo_kind, tfo_cookie) = self._cookie(tcp)
            if tfo_kind is not None:
                rec['tfo_synkind'] = tfo_kind
                rec['tfo_synclen'] = len(tfo_cookie)
                rec['tfo_seq'] = tcp.seq_nbr
                rec['tfo_dlen'] = len(tcp.data) - tcp.doff*4
                rec['tfo_ack'] = 0

        # Look for ACK of TFO data (and cookie)
        elif tcp.syn_flag and tcp.ack_flag and rec['tfo_synkind']:
            rec['tfo_ack'] = tcp.ack_nbr
            (tfo_kind, tfo_cookie) = self._cookie(tcp)
            if tfo_kind is not None:
                rec['tfo_ackkind'] = tfo_kind
                rec['tfo_ackclen'] = len(tfo_cookie)

        # tell observer to keep going
        return True



from scapy.all import IP  # pylint: disable=no-name-in-module
from scapy.all import IPv6  # pylint: disable=no-name-in-module
from scapy.all import UDP  # pylint: disable=no-name-in-module
from scapy.all import TCP  # pylint: disable=no-name-in-module
from scapy.all import DNS  # pylint: disable=no-name-in-module
from scapy.all import DNSQR  # pylint: disable=no-name-in-module
from scapy.all import RandShort  # pylint: disable=no-name-in-module

import pathspider
from pathspider.base import PluggableSpider
from pathspider.forge import ForgeSpider
from pathspider.chains.dns import DNSChain
from pathspider.chains.basic import BasicChain
from pathspider.chains.tcp import TCPChain
from pathspider.chains.evil import EvilChain
from pathspider.chains.base import Chain
from pathspider.chains.tcp import TCP_SYN, TCP_SA

def _evilcon(conn0, conn1):
    if conn0 is True and conn1 is True:
        return 'evilbit.connectivity.works'
    elif conn0 is True and conn1 is False:
        return 'evilbit.connectivity.broken'
    elif conn0 is False and conn1 is True:
        return 'evilbit.connectivity.transient'
    elif conn0 is False and conn1 is False:
        return 'evilbit.connectivity.offline'
 
class EvilBit(ForgeSpider, PluggableSpider):

    name = "evilbit"
    description = "Evil bit connectivity testing"
    version = pathspider.base.__version__
    chains = [BasicChain, DNSChain, TCPChain, EvilChain]
    connect_supported = ["tcpsyn", "dnsudp"]
    packets = 2

    def forge(self, job, seq):
        sport = 0
        while sport < 1024:
            sport = int(RandShort())
        if self.args.connect == 'tcpsyn':
            l4 = (TCP(sport=sport, dport=job['dp']))
        if self.args.connect == 'dnsudp':
            l4 = (UDP(sport=sport, dport=job['dp']) /
                  DNS(qd=DNSQR(qname=job['domain'])))
        if ':' in job['dip']:
            ip = IPv6(src=self.source[1], dst=job['dip'])
        else:
            ip = IP(src=self.source[0], dst=job['dip'])
        if seq == 1:
            ip.flags = 'evil'
        return ip/l4
    

    def combine_flows(self, flows):
        for flow in flows:
            if not flow['observed']:
                return ['pathspider.not_observed']

        conditions = []
        conn0 = False
        conn1 = False

        if self.args.connect == 'tcpsyn':
            if flows[0]['tcp_synflags_rev'] is not None and flows[0][
                    'tcp_synflags_rev'] & TCP_SA == TCP_SA:
                conn0 = True
            if flows[1]['tcp_synflags_rev'] is not None and flows[1][
                    'tcp_synflags_rev'] & TCP_SA == TCP_SA:
                conn1 = True
 
            conditions.append(_evilcon(conn0, conn1))
            if 'works' in (_evilcon(conn0, conn1)):
                if flows[1]['evilbit_syn_rev']:
                    conditions.append('evilbit.mark.seen')
                else:
                    conditions.append('evilbit.mark.not_seen')
                   
        if self.args.connect == 'dnsudp':
            conn0 = flows[0]['dns_response_valid']     
            conn1 = flows[1]['dns_response_valid']     

            conditions.append(_evilcon(conn0, conn1))
            if 'works' in (_evilcon(conn0, conn1)):
                if flows[1]['evilbit_data_rev']:
                    conditions.append('evilbit.mark.seen')
                else:
                    conditions.append('evilbit.mark.not_seen')
        return conditions






from pathspider.tests.chains import ChainTestCase

from pathspider.chains.tfo import TFOChain

class TestTFOChain(ChainTestCase):

    def test_chain_tfo_cookie_request(self):
        test_trace = "tfo_cookie_request.pcap"
        self.create_observer(test_trace, [TFOChain])

        expected_tfo = {
            'tfo_ack': 3865413713,
            'tfo_ackclen': 8,
            'tfo_ackkind': 254,
            'tfo_dlen': 0,
            'tfo_seq': 3865413712,
            'tfo_synclen': 0,
            'tfo_synkind': 254,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        print(flows)
        print(expected_tfo)

        for key in expected_tfo:
            assert flows[0][key] == expected_tfo[key]

    def test_chain_tfo_data_on_syn(self):
        test_trace = "tfo_data_on_syn.pcap"
        self.create_observer(test_trace, [TFOChain])

        expected_tfo = {
            'tfo_ack': 3108342141,
            'tfo_ackclen': 0,
            'tfo_ackkind': 0,
            'tfo_dlen': 39,
            'tfo_seq': 3108342101,
            'tfo_synclen': 8,
            'tfo_synkind': 254,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        print(flows)
        print(expected_tfo)

        for key in expected_tfo:
            assert flows[0][key] == expected_tfo[key]



"""
.. module:: pathspider.chains.base
   :synopsis: An abstract flow observer chain

This module contains the abstract chain class which should be subclassed by
implementations of flow analysis chains for PATHspider's Observer.

.. codeauthor:: Iain R. Learmonth <irl@fsfe.org>

"""

class Chain:
    """
    This is an abstract flow analysis chain. It is intended that all flow
    analysis chains will subclass this class and it is not intended for this
    class to be directly used by PATHspider plugins.
    """

    def new_flow(self, rec, ip):
        """
        This function is called for every new flow to initialise a flow record
        with the fields that will be used by this chain. It is recommended to
        initialise all fields to None until other functions have set values for
        them to make clear which fields are set by this chain and to avoid key
        errors later.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IP or IPv6 packet that triggered the creation of a new
                   flow record
        :type ip: plt.ip or plt.ip6
        """

        raise NotImplementedError("Cannot create a new flow with an abstract Observer chain")



"""
.. module:: pathspider.chains.dns
   :synopsis: A flow analysis chain for the Domain Name System

This module contains the DNSChain flow analysis chain which can be used by
PATHspider's Observer for recording Domain Name System [RFC1035]_ details.

.. codeauthor:: Iain R. Learmonth <irl@fsfe.org>

"""

from pathspider.chains.base import Chain

class DNSChain(Chain):
    """
    This flow analysis chain records details from Domain Name System
    application data.

    +------------------------+------+-----------------------------------------+
    | Field Name             | Type | Meaning                                 |
    +========================+======+=========================================+
    | ``dns_response_valid`` | bool | The flow contained a valid DNS response |
    +------------------------+------+-----------------------------------------+
    """

    def new_flow(self, rec, ip):
        """
        For a new flow, all fields will be initialised to ``False``.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IP or IPv6 packet that triggered the creation of a new
                   flow record
        :type ip: plt.ip or plt.ip6
        :return: Always ``True``
        :rtype: bool
        """

        rec['dns_response_valid'] = False

        return True

    def tcp(self, rec, tcp, rev):
        """
        Records DNS details from TCP segment.

        DNS Response
            If the packet contains a payload, an attempt is made to parse it
            and if successful the ``dns_response_valid`` field is set to ``True``
            if it was a response (not a query).

        :param rec: the flow record
        :type rec: dict
        :param tcp: the TCP packet that was observed to be part of this flow
        :type ip: plt.tcp
        :param rev: ``True`` if the packet was in the reverse direction, ``False`` if
                    in the forward direction
        :type rev: bool
        :return: ``False`` if a valid DNS response has been seen, otherwise ``True``
        :rtype: bool
        """

        if tcp.payload is not None:
            return self._dns_response(rec, tcp.payload, rev)
        else:
            return True

    def udp(self, rec, udp, rev):
        """
        Records DNS details from UDP datagram.

        DNS Response
            If the packet contains a payload, an attempt is made to parse it
            and if successful the ``dns_response_valid`` field is set to ``True``
            if it was a response (not a query).

        :param rec: the flow record
        :type rec: dict
        :param tcp: the UDP packet that was observed to be part of this flow
        :type ip: plt.udp
        :param rev: ``True`` if the packet was in the reverse direction, ``False`` if
                    in the forward direction
        :type rev: bool
        :return: ``False`` if a valid DNS response has been seen, otherwise ``True``
        :rtype: bool
        """

        return self._dns_response(rec, udp.payload, rev)

    def _dns_response(self, rec, payload, rev):
        try:
            from pldns import ldns # pylint: disable=E0611

            if rev is True:
                dns = ldns(payload)
                if dns.is_ok():
                    if dns.is_response:
                        rec['dns_response_valid'] = True
        except ImportError:
            raise RuntimeError("python-libtrace is not installed! "
                               "Cannot dissect DNS!")
        except ValueError:
            pass # Wasn't a DNS payload
        return not rec['dns_response_valid']




import socket

from pathspider.base import CONN_OK
from pathspider.base import CONN_TIMEOUT
from pathspider.base import CONN_FAILED

def connect_tcp(source, job, conn_timeout, sockopts=None):
    """
    This helper function will perform a TCP connection. It will not perform
    any special action in the event that this is the experimental flow,
    it only performs a TCP connection.
    """

    if sockopts is None:
        sockopts = []

    if not isinstance(conn_timeout, int):
        raise RuntimeError("Plugin did not set TCP connect conn_timeout.")

    try:
        if ":" in job['dip']:
            sock = socket.socket(socket.AF_INET6)
            sock.bind((source[1], 0))
        else:
            sock = socket.socket(socket.AF_INET)
            sock.bind((source[0], 0))

        for o in sockopts:
            sock.setsockopt(*o)

        sock.settimeout(conn_timeout)
        sock.connect((job['dip'], job['dp']))

        sp = sock.getsockname()[1]

        sock.shutdown(socket.SHUT_RDWR)
        sock.close()

        return {'sp': sp, 'spdr_state': CONN_OK}
    except TimeoutError:
        return {'sp': sock.getsockname()[1], 'spdr_state': CONN_TIMEOUT}
    except TypeError: # Caused by not having a v4/v6 address when trying to bind
        return {'sp': 0, 'spdr_state': CONN_FAILED}
    except OSError:
        return {'sp': 0, 'spdr_state': CONN_FAILED}



"""
.. module:: pathspider.chains.noop
   :synopsis: A flow observer chain that does nothing

This module contains the NoOpChain flow analysis chain which can be used by
PATHspider's Observer.

.. codeauthor:: Iain R. Learmonth <irl@fsfe.org>

"""

from pathspider.chains.base import Chain

class NoOpChain(Chain):
    """
    This flow analysis chain does not perform any analysis and is present here
    for the purpose of documentation and testing.
    """

    def new_flow(self, rec, ip): # pylint: disable=unused-argument
        """
        This function is called for every new flow to initialise a flow record
        with the fields that will be used by this chain. It is recommended to
        initialise all fields to None until other functions have set values for
        them to make clear which fields are set by this chain and to avoid key
        errors later.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IP or IPv6 packet that triggered the creation of a new
                   flow record
        :type ip: plt.ip or plt.ip6
        :return: True if flow should be kept, False if flow should be discarded
        :rtype: bool
        """
        return True

    def ip4(self, rec, ip, rev): # pylint: disable=unused-argument
        """
        This function is called for every new IPv4 packet seen. It can be used
        to record details for fields present in the IPv4 header.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IPv4 packet that was observed to be part of this flow
        :type ip: plt.ip
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: True if flow should continue to be observed, False if the flow
                 should be passed on for merging (i.e. the flow is complete)
        :rtype: bool
        """

        return True

    def ip6(self, rec, ip6, rev): # pylint: disable=unused-argument
        """
        This function is called for every new IPv6 packet seen. It can be used
        to record details for fields present in the IPv6 header.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IPv6 packet that was observed to be part of this flow
        :type ip: plt.ip6
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: True if flow should continue to be observed, False if the flow
                 should be passed on for merging (i.e. the flow is complete)
        :rtype: bool
        """

        return True

    def icmp4(self, rec, ip, q, rev): # pylint: disable=unused-argument
        """
        This function is called for every new ICMPv4 packet seen. It can be
        used to record details for fields present in the ICMPv4 header or
        quotation.

        .. note:: The IP header is passed as the argument, not the ICMP header
                  as it may be desirable to access fields in the IP header, for
                  instance to determine the router or host that sent the ICMP
                  message

        :param rec: the flow record
        :type rec: dict
        :param ip: the IPv4 packet that was observed to be part of this flow
                   and contained an ICMPv4 header
        :type ip: plt.ip
        :param q: the ICMP quotation of the packet that triggered this message
                  (if any)
        :type q: plt.ip
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: True if flow should continue to be observed, False if the flow
                 should be passed on for merging (i.e. the flow is complete)
        :rtype: bool
        """

        return True

    def icmp6(self, rec, ip6, q, rev): # pylint: disable=unused-argument
        """
        This function is called for every new ICMPv6 packet seen. It can be
        used to record details for fields present in the ICMPv6 header or
        quotation.

        .. note:: The IP header is passed as the argument, not the ICMP header
                  as it may be desirable to access fields in the IP header, for
                  instance to determine the router or host that sent the ICMP
                  message

        :param rec: the flow record
        :type rec: dict
        :param ip: the IPv6 packet that was observed to be part of this flow
                   and contained an ICMPv6 header
        :type ip: plt.ip6
        :param q: the ICMP quotation of the packet that triggered this message
                  (if any)
        :type q: plt.ip6
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: True if flow should continue to be observed, False if the flow
                 should be passed on for merging (i.e. the flow is complete)
        :rtype: bool
        """

        return True

    def tcp(self, rec, tcp, rev): # pylint: disable=unused-argument
        """
        This function is called for every new TCP packet seen. It can be used
        to record details for fields present in the TCP header.

        :param rec: the flow record
        :type rec: dict
        :param tcp: the TCP segment that was observed to be part of this flow
        :type ip: plt.tcp
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: True if flow should continue to be observed, False if the flow
                 should be passed on for merging (i.e. the flow is complete)
        :rtype: bool
        """

        return True

    def udp(self, rec, udp, rev): # pylint: disable=unused-argument
        """
        This function is called for every new UDP packet seen. It can be used
        to record details for fields present in the UDP header.

        :param rec: the flow record
        :type rec: dict
        :param tcp: the UDP segment that was observed to be part of this flow
        :type ip: plt.udp
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: True if flow should continue to be observed, False if the flow
                 should be passed on for merging (i.e. the flow is complete)
        :rtype: bool
        """

        return True




import argparse
import logging
import json
import queue
import signal
import sys
import threading

from straight.plugin import load

from pathspider.base import SHUTDOWN_SENTINEL
from pathspider.base import QUEUE_SIZE

from pathspider.chains.base import Chain

from pathspider.observer import Observer

from pathspider.network import interface_up

chains = load("pathspider.chains", subclasses=Chain)

def run_observer(args):
    logger = logging.getLogger("pathspider")

    if args.list_chains:
        print("The following chains are available:\n")
        for chain in chains:
            print(chain.__name__.lower()[:-5])
        print("\nSpider safely!")
        sys.exit(0)

    interface = args.interface

    if not ":" in args.interface:
        if interface.endswith(".pcap") or interface.startswith("/"):
            interface = "pcapfile:" + interface
        else:
            interface = "int:" + interface

    if interface.startswith("int:") and not interface_up(interface[4:]):
        logger.error("The chosen interface is not up! Cannot continue.")
        logger.error("Try --help for more information.")
        sys.exit(1)

    logger.info("creating observer...")

    chosen_chains = []
    for chosen_chain in args.chains:
        for chain in chains:
            if chosen_chain.lower() + "chain" == chain.__name__.lower():
                chosen_chains.append(chain)

    if len(args.chains) > len(chosen_chains):
        logger.error("Unable to find one or more of the requested chains.")
        logger.error("Try --list-chains to list the available chains.")

    observer_shutdown_queue = queue.Queue(QUEUE_SIZE)
    flowqueue = queue.Queue(QUEUE_SIZE)

    observer = Observer(interface, chosen_chains)

    logger.info("starting observer...")
    threading.Thread(target=observer.run_flow_enqueuer, args=(flowqueue,observer_shutdown_queue)).start()

    logger.info("opening output file " + args.output)
    with open(args.output, 'w') as outputfile:
        logger.info("registering interrupt...")
        def signal_handler(signal, frame):
            observer_shutdown_queue.put(True)
        signal.signal(signal.SIGINT, signal_handler)
        while True:
            result = flowqueue.get()
            if result == SHUTDOWN_SENTINEL:
                logger.info("output complete")
                break
            outputfile.write(json.dumps(result) + "\n")
            logger.debug("wrote a result")

def register_args(subparsers):
    class SubcommandHelpFormatter(argparse.RawDescriptionHelpFormatter):
        def _format_action(self, action):
            parts = super()._format_action(action)
            if action.nargs == argparse.PARSER:
                parts = "\n".join([line for line in parts.split("\n")[1:]])
                parts += "\n\nSpider safely!"
            return parts

    parser = subparsers.add_parser(name='observe',
                                   help="Passively observe network traffic",
                                   formatter_class=SubcommandHelpFormatter)
    parser.add_argument('--list-chains', help="Prints a list of available chains",
                        action='store_true')
    parser.add_argument('-i', '--interface', default="eth0",
                        help=("The interface to use for the observer. If this "
                              "argument ends with '.pcap' then it will instead "
                              "be treated as a PCAP file for offline analysis. "
                              "(Default: eth0)"))
    parser.add_argument('--output', default='/dev/stdout', metavar='OUTPUTFILE',
                        help=("The file to output results data to. "
                              "Defaults to standard output."))
    parser.add_argument('chains', nargs='*', help="Observer chains to use")

    # Set the command entry point
    parser.set_defaults(cmd=run_observer)




from pathspider.tests.chains import ChainTestCase

from pathspider.chains.dns import DNSChain

class TestDNSChain(ChainTestCase):

    def test_chain_dns_valid_response(self):
        test_trace = "dns_valid_response.pcap"
        self.create_observer(test_trace, [DNSChain])

        expected_dns = {
            'dns_response_valid': True,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_dns:
            assert flows[0][key] == expected_dns[key]

    def test_chain_dns_valid_response_tcp(self):
        test_trace = "dns_valid_response_tcp.pcap"
        self.create_observer(test_trace, [DNSChain])

        expected_dns = {
            'dns_response_valid': True,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        print(flows)

        for key in expected_dns:
            assert flows[0][key] == expected_dns[key]

    def test_chain_dns_no_response(self):
        test_trace = "dns_no_response.pcap"
        self.create_observer(test_trace, [DNSChain])

        expected_dns = {
            'dns_response_valid': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_dns:
            assert flows[0][key] == expected_dns[key]



from pathspider.chains.ecn import ECNChain
from pathspider.plugins.ecn import ECN
from pathspider.tests.chains import ChainTestCase
from pathspider.chains.tcp import TCP_SA
from pathspider.base import CONN_OK
from pathspider.base import CONN_FAILED
from pathspider.chains.tcp import TCP_SAE
from pathspider.chains.tcp import TCP_SAEC
from pathspider.chains.tcp import TCP_SYN

def test_plugin_ecn_not_observed():
    flows = [
             {'observed': True, 'spdr_state': CONN_OK},
             {'observed': False, 'spdr_state': CONN_OK}
            ]
    conditions = ECN.combine_flows(None, flows)
    assert "pathspider.not_observed" in conditions

def test_plugin_ecn_combine():
    test_groups = [
                   (CONN_OK,  CONN_OK,  "ecn.connectivity.works"),
                   (CONN_OK,  CONN_FAILED, "ecn.connectivity.broken"),
                   (CONN_FAILED, CONN_OK,  "ecn.connectivity.transient"),
                   (CONN_FAILED, CONN_FAILED, "ecn.connectivity.offline")
                  ]
    for group in test_groups:
        flow = [
                {'observed': True, 'spdr_state': group[0], 'tcp_connected': False},
                {'observed': True, 'spdr_state': group[1], 'tcp_connected': False}
               ]

        conditions = ECN.combine_flows(None, flow)
        assert group[2] in conditions

    test_groups_ecn = [
                       (True,  TCP_SAE,  "ecn.negotiation.succeeded"),
                       (True,  TCP_SAEC, "ecn.negotiation.reflected"),
                       (True,  TCP_SYN,     "ecn.negotiation.failed"),
                      ]

    for group in test_groups_ecn:
        flow = [
                {'observed': True, 'spdr_state': CONN_OK},
                {'observed': True, 'spdr_state': CONN_OK, 'tcp_connected': group[0], 'tcp_synflags_rev': group[1], 'ecn_ect0_syn_rev': False, 'ecn_ect1_syn_rev': False, 'ecn_ce_syn_rev': False, 'ecn_ect0_data_rev': False, 'ecn_ect1_data_rev': False, 'ecn_ce_data_rev': False }
               ]

        conditions = ECN.combine_flows(None, flow)
        assert group[2] in conditions

    test_groups_mark = [
                       (True, "ecn.ipmark.ect0.seen"),
                       (False, "ecn.ipmark.ect0.not_seen"),
                       (True, "ecn.ipmark.ect1.seen"),
                       (False, "ecn.ipmark.ect1.not_seen"),
                       (True, "ecn.ipmark.ce.seen"),
                       (False, "ecn.ipmark.ce.not_seen")
                      ]
    for group in test_groups_mark:
        flow = [
                {'observed': True, 'spdr_state': CONN_OK},
                {'observed': True, 'spdr_state': CONN_OK, 'tcp_connected': True, 'tcp_synflags_rev': TCP_SAE, 'ecn_ect0_syn_rev': group[0], 'ecn_ect1_syn_rev': group[0], 'ecn_ce_syn_rev': group[0], 'ecn_ect0_data_rev': group[0], 'ecn_ect1_data_rev': group[0], 'ecn_ce_data_rev': group[0] }
               ]
        conditions = ECN.combine_flows(None, flow)
        assert group[1] in conditions




from scapy.all import IP         # pylint: disable=E0611
from scapy.all import IPv6       # pylint: disable=E0611
from scapy.all import UDP        # pylint: disable=E0611
from scapy.all import DNS        # pylint: disable=E0611
from scapy.all import DNSQR      # pylint: disable=E0611
from scapy.all import RandShort  # pylint: disable=E0611

import pathspider.base
from pathspider.base import PluggableSpider
from pathspider.forge import ForgeSpider
from pathspider.chains.basic import BasicChain
from pathspider.chains.dns import DNSChain

class UDPZero(ForgeSpider, PluggableSpider):

    name = "udpzero"
    description = "UDP Zero Checksum"
    version = pathspider.base.__version__
    chains = [BasicChain, DNSChain]
    packets = 2

    def forge(self, job, config):
        sport = 0
        while sport < 1024:
            sport = int(RandShort())
        udp = (UDP(sport=sport, dport=job['dp'])/
               DNS(qd=DNSQR(qname=job['domain'])))
        if ':' in job['dip']:
            ip = IPv6(src=self.source[1], dst=job['dip'])
        else:
            ip = IP(src=self.source[0], dst=job['dip'])
        if config == 1:
            udp.chksum = 0 # If not initialised, Scapy will calculate
        return ip/udp

    def combine_flows(self, flows):
        for flow in flows:
            if not flow['observed']:
                return ['pathspider.not_observed']

        if flows[0]['dns_response_valid'] and flows[1]['dns_response_valid']:
            return ['udpzero.connectivity.works']
        if flows[0]['dns_response_valid'] and not flows[1]['dns_response_valid']:
            return ['udpzero.connectivity.broken']
        if not flows[0]['dns_response_valid'] and flows[1]['dns_response_valid']:
            return ['udpzero.connectivity.transient']
        else:
            return ['udpzero.connectivity.offline']



import pycurl

from pathspider.base import CONN_OK
from pathspider.base import CONN_FAILED
from pathspider.plugins.h2 import H2
from pathspider.helpers.http import connect_http
from pathspider.helpers.http import connect_https
from pathspider.chains.basic import BasicChain
from pathspider.chains.tcp import TCPChain
from pathspider.tests.chains import ChainTestCase

def test_plugin_h2_combine():
    test_groups = [
                   (CONN_OK, CONN_OK,   "h2.connectivity.works"),
                   (CONN_FAILED, CONN_OK,  "h2.connectivity.transient"),
                   (CONN_OK, CONN_FAILED,  "h2.connectivity.broken"),
                   (CONN_FAILED, CONN_FAILED,  "h2.connectivity.offline")
                  ]
    for group in test_groups:
        flows = [              
                 {'spdr_state': group[0]},
                 {'spdr_state': group[1], 'http_info': {pycurl.INFO_HTTP_VERSION: pycurl.CURL_HTTP_VERSION_2_0}}
                ] 
        conditions = H2.combine_flows(None, flows)
        assert group[2] in conditions

    upgrade_groups = [ 
                      ('h2.upgrade.success', pycurl.CURL_HTTP_VERSION_2_0),
                      ('h2.upgrade.failed', None),
                     ]
    for group in upgrade_groups:
        flow =  [
                 {'spdr_state': CONN_OK},
                 {'spdr_state': CONN_OK, 'http_info': {pycurl.INFO_HTTP_VERSION : group[1]}}
                ]
        conditions = H2.combine_flows(None, flow)
        assert group[0] in conditions




from pathspider.tests.chains import ChainTestCase

from pathspider.chains.tcp import TCPChain
from pathspider.chains.tcp import TCP_SYN
from pathspider.chains.tcp import TCP_ACK
from pathspider.chains.tcp import TCP_ECE
from pathspider.chains.tcp import TCP_CWR

class TestTCPChain(ChainTestCase):

    def test_chain_tcp_ipv4_simple(self):
        test_trace = "tcp_ipv4_simple.pcap"
        self.create_observer(test_trace, [TCPChain])

        expected_tcp = {
            'tcp_synflags_fwd': TCP_SYN,
            'tcp_synflags_rev': TCP_SYN | TCP_ACK,
            'tcp_connected': True,
            'tcp_fin_fwd': True,
            'tcp_fin_rev': True,
            'tcp_rst_fwd': False,
            'tcp_rst_rev': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_tcp:
            assert flows[0][key] == expected_tcp[key]

    def test_chain_tcp_ipv6_simple(self):
        test_trace = "tcp_ipv6_simple.pcap"
        self.create_observer(test_trace, [TCPChain])

        expected_tcp = {
            'tcp_synflags_fwd': TCP_SYN,
            'tcp_synflags_rev': TCP_SYN | TCP_ACK,
            'tcp_connected': True,
            'tcp_fin_fwd': True,
            'tcp_fin_rev': True,
            'tcp_rst_fwd': False,
            'tcp_rst_rev': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_tcp:
            assert flows[0][key] == expected_tcp[key]

    def test_chain_tcp_ipv4_ecn(self):
        test_trace = "tcp_ipv4_ecn.pcap"
        self.create_observer(test_trace, [TCPChain])

        expected_tcp = {
            'tcp_synflags_fwd': TCP_SYN | TCP_ECE | TCP_CWR,
            'tcp_synflags_rev': TCP_SYN | TCP_ACK | TCP_ECE,
            'tcp_connected': True,
            'tcp_fin_fwd': True,
            'tcp_fin_rev': True,
            'tcp_rst_fwd': False,
            'tcp_rst_rev': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_tcp:
            assert flows[0][key] == expected_tcp[key]

    def test_chain_tcp_ipv6_ecn(self):
        test_trace = "tcp_ipv6_ecn.pcap"
        self.create_observer(test_trace, [TCPChain])

        expected_tcp = {
            'tcp_synflags_fwd': TCP_SYN | TCP_ECE | TCP_CWR,
            'tcp_synflags_rev': TCP_SYN | TCP_ACK | TCP_ECE,
            'tcp_connected': True,
            'tcp_fin_fwd': True,
            'tcp_fin_rev': True,
            'tcp_rst_fwd': False,
            'tcp_rst_rev': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_tcp:
            assert flows[0][key] == expected_tcp[key]

    def test_chain_tcp_ipv4_rst(self):
        test_trace = "tcp_ipv4_rst.pcap"
        self.create_observer(test_trace, [TCPChain])

        expected_tcp = {
            'tcp_synflags_fwd': TCP_SYN,
            'tcp_synflags_rev': TCP_SYN | TCP_ACK,
            'tcp_connected': True,
            'tcp_fin_fwd': False,
            'tcp_fin_rev': False,
            'tcp_rst_fwd': False,
            'tcp_rst_rev': True,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_tcp:
            assert flows[0][key] == expected_tcp[key]

    def test_chain_tcp_ipv4_only_rst(self):
        test_trace = "tcp_ipv4_only_rst.pcap"
        self.create_observer(test_trace, [TCPChain])

        expected_tcp = {
            'tcp_synflags_fwd': None,
            'tcp_synflags_rev': None,
            'tcp_connected': False,
            'tcp_fin_fwd': False,
            'tcp_fin_rev': False,
            'tcp_rst_fwd': True,
            'tcp_rst_rev': False,
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_tcp:
            assert flows[0][key] == expected_tcp[key]



import struct
import socket

from dnslib.dns import DNSError, DNSRecord, DNSQuestion, QTYPE
from scapy.all import RandShort

from pathspider.base import CONN_OK
from pathspider.base import CONN_TIMEOUT
from pathspider.base import CONN_FAILED


class PSDNSRecord(DNSRecord):
    def spider_send(self, source, job, conn_timeout, tcp=False):
        """
        Send packet to nameserver and return response and source port.
        """
        data = self.pack()
        if ':' in job['dip']:
            inet = socket.AF_INET6
        else:
            inet = socket.AF_INET
        if tcp:
            if len(data) > 65535:
                raise ValueError("Packet length too long: %d" % len(data))
            data = struct.pack("!H", len(data)) + data
            sock = socket.socket(inet, socket.SOCK_STREAM)
            if ':' in job['dip']:
                sock.bind((source[1], 0))
            else:
                sock.bind((source[0], 0))
            sock.settimeout(conn_timeout)
            sock.connect((job['dip'], job['dp']))
            sock.sendall(data)
            sp = sock.getsockname()[1]
            response = None
            try:
                response = sock.recv(8192)
                length = struct.unpack("!H",bytes(response[:2]))[0]
                while len(response) - 2 < length:
                    response += sock.recv(8192)
            except socket.timeout:
                pass
            if response is not None and len(response) > 2:
                try:
                    response = response[2:]
                    PSDNSRecord().parse(response)
                except DNSError:
                    response = None
            sock.close()
        else:
            sp = RandShort()
            sock = socket.socket(inet, socket.SOCK_DGRAM)
            if ':' in job['dip']:
                sock.bind((source[1], sp))
            else:
                sock.bind((source[0], sp))
            sp = sock.getsockname()[1]
            sock.settimeout(conn_timeout)
            sock.sendto(self.pack(), (job['dip'], job['dp']))
            response = None
            try:
                response, server = sock.recvfrom(8192)
            except socket.timeout:
                pass
            if response is not None:
                try:
                    PSDNSRecord().parse(response)
                except DNSError:
                    response = None
            sock.close()
        return (response, sp)


def connect_dns_tcp(source, job, conn_timeout):
    """
    This helper function will perform a DNS query over a TCP connection. It
    will not perform any special action in the event that this is the
    experimental flow, it only performs a DNS query connection.
    """

    return connect_dns(source, job, conn_timeout, tcp=True)

def connect_dns_udp(source, job, conn_timeout):
    """
    This helper function will perform a DNS query over a TCP connection. It
    will not perform any special action in the event that this is the
    experimental flow, it only performs a DNS query connection.
    """

    return connect_dns(source, job, conn_timeout, tcp=False)

def connect_dns(source, job, conn_timeout, tcp=False):
    """
    This helper function will perform a DNS query over a TCP connection. It
    will not perform any special action in the event that this is the
    experimental flow, it only performs a DNS query connection.
    """

    try:
        q = PSDNSRecord(q=DNSQuestion(job['domain'], QTYPE.A))
        response, sp = q.spider_send(source, job, conn_timeout, tcp=tcp)
        if response is None:
            return {'sp': sp, 'spdr_state': CONN_FAILED}
        return {'sp': sp, 'spdr_state': CONN_OK}
    except TimeoutError:
        return {'sp': 0, 'spdr_state': CONN_TIMEOUT}
    except TypeError:  # Caused by not having a v4/v6 address when trying to bind
        return {'sp': 0, 'spdr_state': CONN_FAILED}
    except OSError:
        return {'sp': 0, 'spdr_state': CONN_FAILED}
    except ValueError: # Caused by domain names that don't fit in a DNS query (this should never happen)
        return {'sp': 0, 'spdr_state': CONN_FAILED}




from collections import namedtuple
from tempfile import NamedTemporaryFile

from scapy.all import wrpcap

from pathspider.chains.udp import UDPChain
from pathspider.plugins.udpzero import UDPZero
from pathspider.tests.chains import ChainTestCase

TestArgs = namedtuple('TestArgs', ['connect'])
job = {'dip': '192.0.2.4', 'dp': 80, 'domain': 'example.com'}

class TestPluginUDPZeroForgeObserve(ChainTestCase):

    def test_plugin_udpzero_forge_observer(self):
        spider = UDPZero(0, "", TestArgs(connect="dnsudp"))
        packets = []

        for seq in range(0, spider.packets):
            packets.append(spider.forge(job, seq))
        assert len(packets) == 2

        expected_flows = [
          {
            'udp_zero_checksum_fwd': False,
            'udp_zero_checksum_rev': None
          }, {
            'udp_zero_checksum_fwd': True,
            'udp_zero_checksum_rev': None
          }
        ]

        with NamedTemporaryFile() as test_trace:
            for idx in range(0, 2):
                wrpcap(test_trace.name, [packets[idx]])
                self.create_observer(test_trace.name, [UDPChain])

                flows = self.run_observer()
                assert len(flows) == 1

                for key in expected_flows[idx]:
                    print(key + ">>" + str(flows[0][key]) + ":" +
                          str(expected_flows[0][key]))
                    assert flows[0][key] == expected_flows[idx][key]

def test_plugin_udpzero_forge_diff():
    spider = UDPZero(0, "", TestArgs(connect="dnsudp"))
    packets = []

    for seq in range(0, spider.packets):
        packets.append(spider.forge(job, seq))
    assert len(packets) == 2

    packets[0].payload.chksum = 0
    packets[0].payload.sport = packets[1].payload.sport

    print(packets[0].summary())
    print(packets[1].summary())

    assert bytes(packets[0]) == bytes(packets[1])

def test_plugin_udpzero_combine():
    test_groups = [
                   (True,  True,  "udpzero.connectivity.works"),
                   (True,  False, "udpzero.connectivity.broken"),
                   (False, True,  "udpzero.connectivity.transient"),
                   (False, False, "udpzero.connectivity.offline")
                  ]
    for group in test_groups:
        flows = [
                 {'observed': True, 'dns_response_valid': group[0]},
                 {'observed': True, 'dns_response_valid': group[1]}
                ]
        conditions = UDPZero.combine_flows(None, flows)
        assert group[2] in conditions

def test_plugin_udpzero_combine_not_observed():
    for valid in [True, False]:
        flows = [
                 {'observed': True, 'dns_response_valid': valid},
                 {'observed': False}
                ]
        conditions = UDPZero.combine_flows(None, flows)
        assert "pathspider.not_observed" in conditions




from pathspider.tests.chains import ChainTestCase

from pathspider.chains.udp import UDPChain

class TestUDPChain(ChainTestCase):

    def test_chain_udp_zerochecksum(self):
        test_trace = "udp_zerochecksum.pcap"
        self.create_observer(test_trace, [UDPChain])

        expected_mss = {
            'udp_zero_checksum_fwd': True,
            'udp_zero_checksum_rev': True
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_mss:
            assert flows[0][key] == expected_mss[key]

    def test_chain_udp_nonzerochecksum(self):
        test_trace = "udp_nonzerochecksum.pcap"
        self.create_observer(test_trace, [UDPChain])

        expected_mss = {
            'udp_zero_checksum_fwd': False,
            'udp_zero_checksum_rev': False
        }

        flows = self.run_observer()
        assert len(flows) == 1

        for key in expected_mss:
            assert flows[0][key] == expected_mss[key]




import logging
import json

from pathspider.cmd.measure import job_feeder_ndjson
from pathspider.cmd.measure import job_feeder_csv

class FilterSpider:

    def __init__(self, dp=None):
        self.dp = dp

    def add_job(self, job):
        if self.dp is not None:
            job['dp'] = self.dp
        print(json.dumps(job))

    def shutdown(self):
        pass

def filter(args):
    if args.csv_input:
        job_feeder = job_feeder_csv
    else:
        job_feeder = job_feeder_ndjson

    spider = FilterSpider(dp=args.dp)

    job_feeder("/dev/stdin", spider)

def register_args(subparsers):
    parser = subparsers.add_parser(name='filter',
                                   help="Pre-process a target list")
    parser.add_argument('--csv-input', action='store_true',
                        help=("Indicate CSV format."))
    parser.add_argument('--dp', type=int, default=None,
                        help=("A destination port to add to the targets."))

    # Set the command entry point
    parser.set_defaults(cmd=filter)



from pathspider.desync import DesynchronizedSpider
from pathspider.sync import SynchronizedSpider

class SingleSpider(DesynchronizedSpider):
    # pylint: disable=W0223

    connections = [SynchronizedSpider.connect]
